#author: a4chitect
#version: 1.2
#history
#1.01 tried to fix it to work inside nested groups
#1.02 fixed to work inside nested groups
#1.1 fixed to work with different units and geometry scenarios + welding at cpoint in the selection
#1.12 added 'weld to lowest' point, finds the point which has the lowest z coordinate and moves all other points to this position
#1.2 visual selection of the target based on user (key LEFT/RIGHT) feedback
#possible limits -> working inside groups, nested components
#credits:   whaat (for huge part of this script : screen border drawing)

#licence: your only duty is to share any improvements and/or upgrades of this code under this same condition

require 'sketchup.rb'
#Region Selection Tool# -> SPECIAL THANKS TO WHAAT FOR THIS GREAT MODULE
class SelectRegionTool

def initialize
    @ip1 = nil
    @ip2 = nil
    @xdown = 0
    @ydown = 0
	@x1=0
	@x2=0
	@y1=0
	@y2=0
	
	if(Sketchup.active_model.selection.length == 0)
		UI.messagebox "Please select geometry prior to running this script"
		Sketchup.active_model.select_tool Select
		return
	end
		
	@cam = Sketchup.active_model.active_view.camera
	if(@cam.perspective? == false)
		@ortho = 1
		@cam.perspective = true
	else
		@ortho = 0
	end
	
	@weldfull = 0
end

def activate
    @ip1 = Sketchup::InputPoint.new
    @ip2 = Sketchup::InputPoint.new
    @ip = Sketchup::InputPoint.new
    @drawn = false    
    self.reset(nil)
end

def deactivate(view)
    view.invalidate if @drawn
		if(@ortho == 1)
			@cam.perspective = false
		end
end

def onMouseMove(flags, x, y, view)
    if( @state == 0 )
        # We are getting the first point of the region.  Call the pick method
        # on the InputPoint to get a 3D position from the 2D screen position
        # that is bassed as an argument to this method.
        @ip.pick view, x, y
        if( @ip != @ip1 )
            # if the point has changed from the last one we got, then
            # see if we need to display the point.  We need to display it
            # if it has a display representation or if the previous point
            # was displayed.  The invalidate method on the view is used
            # to tell the view that something has changed so that you need
            # to refresh the view.
            view.invalidate if( @ip.display? or @ip1.display? )
            @ip1.copy! @ip
		@x1=x
		@y1=y
            view.tooltip = @ip1.tooltip
        end
    else
        # Getting the second point of the region
        @ip2.pick view, x, y
        view.tooltip = @ip2.tooltip if( @ip2.valid? )
        view.invalidate
        
        # Update the length displayed in the VCB
        if( @ip2.valid? )
            length = @ip1.position.distance(@ip2.position)
            Sketchup::set_status_text length.to_s, SB_VCB_VALUE
		@x2=x
		@y2=y
        end
        
        # Check to see if the mouse was moved far enough to create a line.
        # This is used so that you can create a line by either draggin
        # or doing click-move-click
        if( (x-@xdown).abs > 10 || (y-@ydown).abs > 10 )
            @dragging = true
        end
    end
		if( @weldfull == 1 && $weldtype == 3)
			view.invalidate
		end
end

# The onLButtonDOwn method is called when the user presses the left mouse button.
def onLButtonDown(flags, x, y, view)
    # When the user clicks the first time, we switch to getting the
    # second point.  When they click a second time we create the line
    if( @state == 0 )
        @ip1.pick view, x, y
        if( @ip1.valid? )
            @state = 1
            Sketchup::set_status_text $exStrings.GetString("Select region"), SB_PROMPT
            @xdown = x
            @ydown = y
		@x1=x
		@y1=y
        end
    else
        # create the line on the second click
        if( @ip2.valid? )
            self.save_region(view)
            self.reset(view)
        end
    end
    
    # Clear any inference lock
    view.lock_inference
end

# The onLButtonUp method is called when the user releases the left mouse button.
def onLButtonUp(flags, x, y, view)
    # If we are doing a drag, then create the line on the mouse up event
    if( @dragging && @ip2.valid? )
        self.save_region(view)
        self.reset(view)
    end
end

# onKeyDown is called when the user presses a key on the keyboard.
# We are checking it here to see if the user pressed the shift key
# so that we can do inference locking
def onKeyDown(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
        @shift_down_time = Time.now
        
        # if we already have an inference lock, then unlock it
        if( view.inference_locked? )
            # calling lock_inference with no arguments actually unlocks
            view.lock_inference
        elsif( @state == 0 && @ip1.valid? )
            view.lock_inference @ip1
        elsif( @state == 1 && @ip2.valid? )
            view.lock_inference @ip2, @ip1
        end
    end
		#if we already have points to weld in an array and we want to specify the target visually
		if( @weldfull == 1 && $weldtype == 3 && (key == 39 || key == 63235)) #39 - right arrow, 63235 on a MAC
			self.cycleweld(1, view)
			view.invalidate
		end
		if( @weldfull == 1 && $weldtype == 3 && (key == 37 || key == 63234)) #37 - left arrow, 63234 on a MAC
			self.cycleweld(-1, view)
			view.invalidate
		end
		#and now we want to make the weld - confirm with enter
		if( @weldfull == 1 && $weldtype == 3 && key == 13 ) #13 - enter key
			@cycled = 1
			self.save_region(view)
			@cycled = 0
			@weldfull = 0
			view.invalidate
		end	
end

# onKeyUp is called when the user releases the key
# We use this to unlock the interence
# If the user holds down the shift key for more than 1/2 second, then we
# unlock the inference on the release.  Otherwise, the user presses shift
# once to lock and a second time to unlock.
def onKeyUp(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY &&
        view.inference_locked? &&
        (Time.now - @shift_down_time) > 0.5 )
        view.lock_inference
    end
end


# The draw method is called whenever the view is refreshed.  It lets the
# tool draw any temporary geometry that it needs to.
def draw(view)
    if( @ip1.valid? )
        if( @ip1.display? )
            @ip1.draw(view)
            @drawn = true
        end
        
        if( @ip2.valid? )
            @ip2.draw(view) if( @ip2.display? )
            
            # The set_color_from_line method determines what color
            # to use to draw a line based on its direction.  For example
            # red, green or blue.
            view.set_color_from_line(@ip1, @ip2)
            self.draw_geometry([@x1,@y1,0],[@x2,@y1,0],[@x2,@y2,0],[@x1,@y2,0], view)
            @drawn = true
        end
    end
		if( @weldfull == 1 && $weldtype == 3)
			self.draw_point(@finalvert,view)
		end
end

# onCancel is called when the user hits the escape key
def onCancel(flag, view)
    self.reset(view)
end


# The following methods are not directly called from SketchUp.  They are
# internal methods that are used to support the other methods in this class.

# Reset the tool back to its initial state
def reset(view)
    # This variable keeps track of which point we are currently getting
    @state = 0
		@x1=0
		@x2=0
		@y1=0
		@y2=0
    
		# Display a prompt on the status bar
		if(@weldfull == 0)
	    Sketchup::set_status_text("Select region", SB_PROMPT)
		elsif($weldtype == 3)
			Sketchup::set_status_text("Cycle through weld targets with LEFT/RIGHT keys, confirm with ENTER", SB_PROMPT)
		end
    
    # clear the InputPoints
    @ip1.clear
    @ip2.clear
    
    if( view )
        view.tooltip = nil
        view.invalidate if @drawn
    end
    
    @drawn = false
    @dragging = false
end

# Create new geometry when the user has selected two points.
def save_region(view)
	if (@weldfull == 0)
		@regselected = true
		#Validate region and store
		min_x=[@x1,@x2].min
		max_x=[@x1,@x2].max
		min_y=[@y1,@y2].min
		max_y=[@y1,@y2].max
		pts = []
		vect = []
		pts.clear
		vect.clear


		
		#UI.messagebox 'minx ' + min_x.to_s + '| maxx ' + max_x.to_s + '| miny ' + min_y.to_s + '| maxy ' + max_y.to_s
		reffacegroup = Sketchup.active_model.active_entities.add_group
		ray1 = view.pickray min_x, min_y
			cam = Geom::Point3d.new ray1[0]
				vect[0] = Geom::Vector3d.new ray1[1]
				pts[0] = cam + vect[0]
		ray2 = view.pickray min_x, max_y
			vect[1] = Geom::Vector3d.new ray2[1]
			pts[1] = cam + vect[1]
		ray3 = view.pickray max_x, max_y
			vect[2] = Geom::Vector3d.new ray3[1]
			pts[2] = cam + vect[2]
		#we have to compute the 4th point differently to assure coplanarity of all points (the above method could produce 4th point not on the same plane as the 3 points before)
		temppoint1 = Geom::Point3d.new pts[0]
		temppoint2 = Geom::Point3d.new pts[1]
		temppoint3 = Geom::Point3d.new pts[2]
		tempplane = Geom.fit_plane_to_points temppoint1, temppoint2, temppoint3
		ray4 = view.pickray max_x, min_y
			vect[3] = Geom::Vector3d.new ray4[1]
			ray4line = [cam, vect[3]]
			pts[3] = Geom.intersect_line_plane(ray4line,tempplane)

		#now we can construct a reference face (reflecting the user selection)
		refface = reffacegroup.entities.add_face pts

		#now lets change the selection to point within the projected boundaries of the refface
		selection_set = Sketchup.active_model.selection
		vertices = []
		@weldvert = []
		weldcpoint = []
		
		sumx = 0
		sumy = 0
		sumz = 0

		selection_set.each {|sc|
			if sc.typename == "Edge" then
				vertices = vertices + sc.vertices
			end
			if sc.typename == "Face" then
				vertices = vertices + sc.vertices
			end
			if sc.typename == "ConstructionPoint" then
				weldcpoint.push sc.position
			end
		}
		vertices.uniq!
		
		if(vertices.length < 2)
			UI.messagebox "Please select at least two vertices"
			self.reset(view)
		end
		0.upto(vertices.length-1) do |i|

			target = Geom::Point3d.new vertices[i]
			direction = Geom::Vector3d.new (target - cam)
			rt = Sketchup.active_model.raytest(cam,direction)
			intersectgroup = rt[1][rt[1].length-2]
	    if ( intersectgroup.to_s==reffacegroup.to_s )
				@weldvert.push vertices[i]
			end
		end
		#now erase the reference plane
		status = reffacegroup.erase!
		@weldfull = 1
		case $weldtype
			when 0
				0.upto(@weldvert.length-1) do |j|
					sumx = sumx + @weldvert[j].position.x
					sumy = sumy + @weldvert[j].position.y
					sumz = sumz + @weldvert[j].position.z
				end
				vertnum = @weldvert.length.to_f
				@finalvert = Geom::Point3d.new(sumx/vertnum,sumy/vertnum,sumz/vertnum)
			when 1
				@finalvert = Geom::Point3d.new weldcpoint[0]
			when 2
				lowestpos = @weldvert[0].position
				0.upto(@weldvert.length-1) do |k|
					if(@weldvert[k].position.z.to_f < lowestpos[2].to_f)
						lowestpos = @weldvert[k].position.to_a
					end
				end
				@finalvert = Geom::Point3d.new lowestpos
			when 3
				#view = view.zoom @weldvert
				@cycleindex = 0
				@finalvert = @weldvert[@cycleindex].position
		end
	end #if @weldfull
	if ($weldtype < 3 || @cycled == 1)
		Sketchup.active_model.start_operation "Weld vertices"
		#move vertices to the final vert (Juan V. Soler's 'brilliant point')
		@weldvert.each{|x|
			#DEBUG      UI.messagebox 'x.position:' + x.position.to_s + '    @finalvert:' + @finalvert.to_s
			t=Geom::Transformation.new(@finalvert-x.position)
			Sketchup.active_model.active_entities.transform_entities(t, x)
	  }
		#now draw a line from the @finalvert to a random position, so when this is erased, all colinear lines are weld together credit for this part goes to Carlo Roosen, for his repair_broken_lines script
		l = Sketchup.active_model.entities.add_line(@finalvert, Geom::Point3d.new(3.45,6.78,9.12))
		l.erase!
		#finish the operation
		Sketchup.active_model.commit_operation
	end
end

def cycleweld(add, view)
	if (@cycleindex == 0 && add == -1)
		@cycleindex = (@weldvert.length - 1)
	elsif (@cycleindex == (@weldvert.length - 1) && add == 1)
		@cycleindex = 0
	else
		@cycleindex = @cycleindex + add
	end
	@finalvert = @weldvert[@cycleindex].position
end

def draw_point(fv, view)
		centerpoint = Geom::Point3d.new fv
		#DEBUG	puts centerpoint
		weldstatus = view.draw_points centerpoint, 10, 7, "yellow"
		xpos = Geom::Vector3d.new(1,0,0)
		ypos = Geom::Vector3d.new(0,1,0)
		zpos = Geom::Vector3d.new(0,0,1)
		x1 = centerpoint - xpos
		x2 = centerpoint + xpos
		y1 = centerpoint - ypos
		y2 = centerpoint + ypos
		z1 = centerpoint - zpos
		z2 = centerpoint + zpos
		view.drawing_color="red"
		view.draw(GL_LINES, [x1,x2])
		view.drawing_color="green"
		view.draw(GL_LINES, [y1,y2])
		view.drawing_color="blue"
		view.draw(GL_LINES, [z1,z2])
end

# Draw the geometry
def draw_geometry(pt1, pt2, pt3, pt4, view)
    view.draw2d(GL_LINE_LOOP,[pt1,pt2,pt3,pt4])
end
end #class

class VW
	def VW::at(weldtype)
		Sketchup.active_model.select_tool SelectRegionTool.new
		$weldtype = weldtype
	end
end #class VW

if( not $vertweld_menu_loaded )
	$vertweld_sub_menu = UI.menu("Plugins").add_submenu("Vertex Weld (From Selection)")
	$vertweld_sub_menu.add_item("Weld at Center") { VW.at(0) }
	$vertweld_sub_menu.add_item("Weld at Construction Point") { VW.at(1) }
	$vertweld_sub_menu.add_item("Weld at Lowest") { VW.at(2) }
	$vertweld_sub_menu.add_item("Weld with Feedback") { VW.at(3) }
$vertweld_menu_loaded = true
end